#------------------------------------------------------------------------------------------------
# KLayout forum No.1642
#
# Sub: Python Scripting for reading an Excel file and creating a polar array on Klayout Macro
# Posted by: dsenuka
#        on: September 16
#
# Author       : Kazzz-S
# Last modified: 2020-09-20
#------------------------------------------------------------------------------------------------
# Usage:
#        (1) Store this script in ${HOME}/.klayout/pymacros/
#        (2) Start KLayout and open a New Layout
#        (3) Set the top cell name 'metalens'
#        (4) Start the Macro Development IDE tool
#        (5) Open this script and run
#        (6) Load (drag & drop) "KL1642.lyp" layer property file      .
#------------------------------------------------------------------------------------------------
import os
import math
import pandas as pd
import pya

#-------------------------------------------------------------------------------------
# Global control parameters; modify and set these parameters as you like
#-------------------------------------------------------------------------------------
Debug          = True   # set False not to print debug messages
Convert2Static = False  # set True to convert PCells to static cells
MyRoot         = os.environ["HOME"] + "/KLayout/Forum1642/"
MyCSV          = MyRoot + "KazzzS-circles.csv"
OutGDS2        = MyRoot + "KL1642.gds"

#-------------------------------------------------------------------------------------
## Compute the number of vertices of a regular-N polygon approximating a circle
#
# @param[in] radius     radius in [um]
# @param[in] tol        tolerance in [um] (default=0.001[um])
# @param[in] integral4  the number should be an integral multiple of four (default=True)
#
# @return the number of vertices
#-------------------------------------------------------------------------------------
def ComputeNumberVerticies( radius, tol=0.001, integral4=True ):
  x = 1.0 - tol / radius
  y = math.acos(x)
  z = int( 2*math.pi / y )
  if not integral4:
    numver = z
  else:
    res = z % 4
    if   res == 0:
      numver = z
    elif res == 1:
      numver = z + 3
    elif res == 2:
      numver = z + 2
    elif res == 3:
      numver = z + 1
    else:
      numver = 0 # dummy
  return numver

#-------------------------------------------------------------------------------------
# KLayout's major objects
#-------------------------------------------------------------------------------------
MainWindow  = pya.MainWindow.instance()
CurrentView = MainWindow.current_view()
Layout      = CurrentView.active_cellview().layout()
DesignFname = CurrentView.active_cellview().filename()
TopCellPtr  = Layout.top_cell()
TopCellName = TopCellPtr.name   # Python getter
MyDBU       = Layout.dbu        # Python getter; database unit; most likely 0.001 [um]

#---------------------------------------------------------------------------------------
## Construct the circle dictionary by reading the (modified) CSV file with four columns
#
#      #------------------------------------------------------------------------------------------------
#      # KLayout forum No.1642
#      #
#      # Sub: Python Scripting for reading an Excel file and creating a polar array on Klayout Macro
#      # Posted by: dsenuka
#      #        on: September 16
#      #
#      # Remade by    : Kazzz-S
#      # Last modified: 2020-09-20
#      #------------------------------------------------------------------------------------------------
#      circleID,orbitRadius[um],circleDiameter[um],numInstance
#      1,0,1.2,1
#      2,1.5,1.195175081,5
#      3,3,1.180720617,12
#      4,4.5,1.156697054,18
#      5,6,1.123203748,25
#      6,7.5,1.083557066,31
#      7,9,1.039992702,37
#      8,10.5,0.991020403,43
#      9,12,0.943702396,50
#      10,13.5,0.892094581,56
#      11,15,0.842634366,62
#      12,16.5,0.783521876,69
#      13,18,0.698041749,75
#      14,19.5,1.114193525,81
#      15,21,1.01447834,87
#---------------------------------------------------------------------------------------
def ConstructCircleDictionary():
  global CircleDic    # the circle dictionary

  CircleDic   = dict()
  df          = pd.read_csv( MyCSV, comment='#' )
  numLeafCell = len(df)
  for idx in range(0, numLeafCell):
    cirID    = int  ( df.iloc[idx,0] )
    orbitRad = round( df.iloc[idx,1], 3 )
    cirDia   = round( df.iloc[idx,2], 4 ) # found too many digits below the decimal point
    numInst  = int  ( df.iloc[idx,3] )
    CircleDic[cirID] = (orbitRad, cirDia, numInst)

  if Debug:
    print( "### Contents of CircleDic ###" )
    for cirID in sorted(CircleDic.keys()):
      print( cirID, CircleDic[cirID] )
    print( "" )

#-------------------------------------------------------------------------------
## Create different circles as PCells and populate the cell dictionary
#-------------------------------------------------------------------------------
def CreateLeafPCells():
  global CellDic    # the cell dictionary

  CellDic = dict()

  #-----------------------------------------------------------------------
  # [1] Create each PCell as "ELLIPSE" ("CIRCLE" should be OK, I guess)
  #-----------------------------------------------------------------------
  for cirID in sorted(CircleDic.keys()):
    #---------------------------------------------------------------------
    # (A) Populate a dictionary to pass PCell parameters
    #---------------------------------------------------------------------
    (orbitRad, cirDia, numInst) = CircleDic[cirID]

    cellName = "circle_%d" % cirID
    numVer   = ComputeNumberVerticies( cirDia / 2.0 )

    paraL  = "layer"           # Layer and Datatype
    paraRX = "radius_x"        # radius_x shown in the dialog for further CAD operation
    paraRY = "radius_y"        # radius_y shown in the dialog for further CAD operation
    paraAX = "actual_radius_x" # hidden parameter that actually changes the radius_x
    paraAY = "actual_radius_y" # hidden parameter that actually changes the radius_y
    paraN  = "npoints"         # number of vertices of a polygon approximating the circle
    """
      Other parameters defined in "libBasicEllipse.cc"
      Set these, too, if required.

          //  parameter #3: x handle
          tl_assert (parameters.size () == p_handle_x);
          parameters.push_back (db::PCellParameterDeclaration ("handle_x"));
          parameters.back ().set_type (db::PCellParameterDeclaration::t_shape);
          parameters.back ().set_default (db::DPoint (-0.2, 0));
          parameters.back ().set_description (tl::to_string (tr ("Rx")));

          //  parameter #4: x handle
          tl_assert (parameters.size () == p_handle_y);
          parameters.push_back (db::PCellParameterDeclaration ("handle_y"));
          parameters.back ().set_type (db::PCellParameterDeclaration::t_shape);
          parameters.back ().set_default (db::DPoint (-0.1, 0));
          parameters.back ().set_description (tl::to_string (tr ("Ry")));

    """
    params = dict()
    params[paraL]  = pya.LayerInfo( cirID, 0 )
    # [2020-09-20]
    # My initial idea was to provide different cells with different layers for better visibility.
    # Of course, you can use a single layer by passing a constant integer instead of 'cirID.'
    # If the number of distinct cells exceeds 1000, it will be wise to avoid too many layers.
    params[paraRX] = cirDia / 2.0
    params[paraRY] = cirDia / 2.0
    params[paraAX] = cirDia / 2.0
    params[paraAY] = cirDia / 2.0
    params[paraN]  = numVer

    #---------------------------------------------------------------------
    # (B) Create a PCell and convert it to a static cell on demand
    #---------------------------------------------------------------------
    sbcell1 = Layout.create_cell( "ELLIPSE", "Basic", params ).cell_index()
    if Convert2Static:
      sbcell2 = Layout.convert_cell_to_static( sbcell1 )
      Layout.rename_cell( sbcell2, cellName )
      Layout.delete_cell( sbcell1 )
      CellDic[cirID] = sbcell2
      if Debug:
        print( "Leaf static cell <%s> has Cell-ID <%d>" % (cellName, sbcell2) )
    else:
      Layout.rename_cell( sbcell1, cellName )
      CellDic[cirID] = sbcell1
      if Debug:
        print( "Leaf PCell <%s> has Cell-ID <%d>" % (cellName, sbcell1) )

  #-----------------------------------------------------------------------
  # [2] Take care of the newly added layers
  #-----------------------------------------------------------------------
  CurrentView.add_missing_layers()  # add missing layers
  CurrentView.zoom_fit()            # zoom to fit the entire hierarchy

#-------------------------------------------------------------------------------
## Create the ring array
#-------------------------------------------------------------------------------
def CreateRingArray():
  for cirID in sorted(CircleDic.keys()):

    #-----------------------------------------------------
    # [1] Look up the dictionaries
    #-----------------------------------------------------
    (orbitRad, cirDia, numInst) = CircleDic[cirID]
    leafCell = CellDic[cirID]

    #-----------------------------------------------------
    # [2] Place the leaf cell one by one
    #-----------------------------------------------------
    incRot = 2*math.pi / numInst
    for idx in range(0, numInst):
      #-----------------------------------------------------
      # (A) Set the location on the orbit
      #-----------------------------------------------------
      theta = idx * incRot
      x     = orbitRad * math.cos(theta)
      y     = orbitRad * math.sin(theta)

      #-----------------------------------------------------
      # (B) SREF the PCell
      #-----------------------------------------------------
      dispVec = pya.DTrans( x/MyDBU, y/MyDBU )  # displace cell by (x, y) [um]
      cellIns = pya.CellInstArray()             # default cell instance
      cellIns.cell_index = leafCell             # 'set' the cell to refer (what to refer)
      cellIns.trans = dispVec                   # 'set' the displacement vector (how to refer)
      TopCellPtr.insert( cellIns )              # insert the instance to [TopCellName]

#-------------------------------------------------------------------------------
## Generate the GDS2 file
#-------------------------------------------------------------------------------
def GenerateGDS2():
  sop = pya.SaveLayoutOptions()   # default save option
  Layout.write( OutGDS2, sop )    # *.gds makes KLayout infer the output format
  print( "" )
  print( "### Generated", OutGDS2 )

#===============================================================================
# Main()
#===============================================================================

#-------------------------------------------------------------
# [1] Construct the cell hierarchy
#-------------------------------------------------------------
ConstructCircleDictionary()
CreateLeafPCells()
CreateRingArray()

#-------------------------------------------------------------
# [2] Change the hierarchy level
#-------------------------------------------------------------
CurrentView.max_hier_levels = 2
CurrentView.zoom_fit()

#-------------------------------------------------------------
# [3] Generate the GDS2 file
#-------------------------------------------------------------
GenerateGDS2()

#-------------------------------------------------------------
# [4] Manual operations can follow...
#-------------------------------------------------------------
# You can
#     (A) change the layer properties
#     (B) convert a PCell to a static cell, if not yet done
#     (C) flat the cell hierarchy
#     (D) change the layers
#     (E) and many more...
#
# Enjoy PYA scripting!
#-------------------------------------------------------------

#---------------
# End of file
#---------------
